"""
title: Entrypoint module, in case you use `python -m arx`.
references:
  - https://www.python.org/dev/peps/pep-0338/
  - https://docs.python.org/2/using/cmdline.html#cmdoption-m
  - https://docs.python.org/3/using/cmdline.html#cmdoption-m
"""

from arx.cli import app

if __name__ == "__main__":
    app()
